@@ -16,7 +16,7 @@ except ImportError:
     from bson.objectid import ObjectId
 
 from dateutil import parser
-from dagobah.backend.base import BaseBackend
+from ..backend.base import BaseBackend
 
 TRUNCATE_LOG_SIZES_CHAR = {'stdout': 500000,
                            'stderr': 500000}
@@ -24,6 +24,10 @@ TRUNCATE_LOG_SIZES_CHAR = {'stdout': 500000,
 
 class MongoBackend(BaseBackend):
     """ Mongo Backend implementation """
+    required_packages = [{'pypi_name': 'pymongo',
+                          'module_name': 'pymongo',
+                          'version_key': 'version',
+                          'version': '2.5'}]
 
     def __init__(self, host, port, db, dagobah_collection='dagobah',
                  job_collection='dagobah_job', log_collection='dagobah_log'):
@@ -135,3 +139,16 @@ class MongoBackend(BaseBackend):
         for rec in cur:
             return rec
         return {}
+
+    def get_run_log_history(self, job_id, task_name, limit=10):
+        q = {'job_id': ObjectId(job_id),
+             'tasks.%s' % task_name: {'$exists': True}}
+        cur = self.log_coll.find(q).sort([('save_date',
+                                           pymongo.DESCENDING)]).limit(limit)
+        return list(cur)
+
+    def get_run_log(self, job_id, task_name, log_id):
+        q = {'job_id': ObjectId(job_id),
+             'tasks.%s' % task_name: {'$exists': True},
+             'log_id': ObjectId(log_id)}
+        return self.log_coll.find_one(q)['tasks'][task_name]
