@@ -6,8 +6,8 @@ import json
 from flask import request, abort, send_file
 from flask_login import login_required
 
-from dagobah.daemon.daemon import app
-from dagobah.daemon.util import validate_dict, api_call, allowed_file
+from .daemon import app
+from .util import validate_dict, api_call, allowed_file
 
 dagobah = app.config['dagobah']
 
@@ -35,6 +35,43 @@ def get_job():
     return job._serialize()
 
 
+@app.route('/api/logs', methods=['GET'])
+@login_required
+@api_call
+def get_run_log_history():
+    args = dict(request.args)
+    if not validate_dict(args,
+                         required=['job_name', 'task_name'],
+                         job_name=str,
+                         task_name=str):
+        abort(400)
+
+    job = dagobah.get_job(args['job_name'])
+    task = job.tasks.get(args['task_name'], None)
+    if not task:
+        abort(400)
+    return task.get_run_log_history()
+
+
+@app.route('/api/log', methods=['GET'])
+@login_required
+@api_call
+def get_log():
+    args = dict(request.args)
+    if not validate_dict(args,
+                         required=['job_name', 'task_name', 'log_id'],
+                         job_name=str,
+                         task_name=str,
+                         log_id=str):
+        abort(400)
+
+    job = dagobah.get_job(args['job_name'])
+    task = job.tasks.get(args['task_name'], None)
+    if not task:
+        abort(400)
+    return task.get_run_log(args['log_id'])
+
+
 @app.route('/api/head', methods=['GET'])
 @login_required
 @api_call
@@ -148,12 +185,14 @@ def add_task_to_job():
                          required=['job_name', 'task_command', 'task_name'],
                          job_name=str,
                          task_command=str,
-                         task_name=str):
+                         task_name=str,
+                         task_target=str):
         abort(400)
 
     dagobah.add_task_to_job(args['job_name'],
                             args['task_command'],
-                            args['task_name'])
+                            args['task_name'],
+                            hostname=args.get("task_target", None))
 
 
 @app.route('/api/delete_task', methods=['POST'])
@@ -188,7 +227,6 @@ def add_dependency():
     job = dagobah.get_job(args['job_name'])
     job.add_dependency(args['from_task_name'], args['to_task_name'])
 
-
 @app.route('/api/delete_dependency', methods=['POST'])
 @login_required
 @api_call
@@ -319,6 +357,21 @@ def edit_job():
     job.edit(**args)
 
 
+@app.route('/api/update_job_notes', methods=['POST'])
+@login_required
+@api_call
+def update_job_notes():
+    args = dict(request.form)
+    if not validate_dict(args,
+                         required=['job_name', 'notes'],
+                         job_name=str,
+                         notes=str):
+        abort(400)
+
+    job = dagobah.get_job(args['job_name'])
+    job.update_job_notes(args['notes'])
+
+
 @app.route('/api/edit_task', methods=['POST'])
 @login_required
 @api_call
@@ -331,7 +384,8 @@ def edit_task():
                          name=str,
                          command=str,
                          soft_timeout=int,
-                         hard_timeout=int):
+                         hard_timeout=int,
+                         hostname=str):
         abort(400)
 
     job = dagobah.get_job(args['job_name'])
@@ -339,6 +393,14 @@ def edit_task():
     if not task:
         abort(400)
 
+    # validate host
+    if args.get('hostname') not in dagobah.get_hosts():
+        # Check for empty host, if so then task is no longer remote
+        if not args.get('hostname'):
+            args['hostname'] = None
+        else:
+            abort(400)
+
     del args['job_name']
     del args['task_name']
     job.edit_task(task.name, **args)
@@ -412,3 +474,10 @@ def import_job():
     file = request.files['file']
     if (file and allowed_file(file.filename, ['json'])):
         dagobah.add_job_from_json(file.read(), destructive=True)
+
+
+@app.route('/api/hosts', methods=['GET'])
+@login_required
+@api_call
+def get_hosts():
+    return dagobah.get_hosts()
